==> conc-2.0.1.tar.gz: conc-2.0.1/common/matrix.c <==
#include "matrix.h"

#define MATRIX_BUFSIZ 2048

typedef struct _Connection {
    guint reader;
    GList *to; /* list of GIOChannel */
} Connection;

struct _Matrix {
    GTree *connections; /* tree of Connection */
};

static void
matrix_tree_free(gpointer data) {
    Connection *c = data;

    if (c->reader > 0) {
	g_source_remove(c->reader);
    }
    g_list_free(c->to);
    g_free(c);
}

static gint
matrix_compare(gconstpointer a,
	gconstpointer b,
	gpointer data) {
    return ((gint)b - (gint)a);
}

static gboolean
matrix_reader(GIOChannel *source,
	GIOCondition condition,
	gpointer data) {
    Connection *c = data;
    gchar buf[MATRIX_BUFSIZ];
    gsize bytes;
    GList *l;
    GError *e = NULL;

    switch (g_io_channel_read_chars(source, buf, MATRIX_BUFSIZ, &bytes, &e)) {
	case G_IO_STATUS_AGAIN:
	    return TRUE;
	case G_IO_STATUS_EOF:
	    return FALSE;
	case G_IO_STATUS_ERROR:
	    g_printerr("error reading for matrix (%s)\n", e->message);
	    g_clear_error(&e);
	    return FALSE;
	case G_IO_STATUS_NORMAL:
	    break;
	default:
	    return TRUE;
    }
    for (l = c->to; l; l = g_list_next(l)) {
	GIOChannel *to = l->data;

	g_io_channel_write_chars(to, buf, bytes, NULL, NULL);
    }
    return TRUE;
}

extern Matrix
matrix_new(void) {
    Matrix m = g_new0(struct _Matrix, 1);

    m->connections = g_tree_new_full(matrix_compare, NULL, NULL,
	    matrix_tree_free);
    return m;
}

extern gboolean
matrix_connect(Matrix matrix,
	GIOChannel *from,
	GIOChannel *to,
	GError **error) {
    Connection *c;

    g_assert(matrix);
    g_assert(from);
    g_assert(to);
    c = g_tree_lookup(matrix->connections, from);
    if (!c) {
	c = g_new0(Connection, 1);
	c->reader = g_io_add_watch(from, G_IO_IN, matrix_reader, c);
	g_tree_insert(matrix->connections, from, c);
    }
    if (g_list_find(c->to, to) == NULL) {
	c->to = g_list_prepend(c->to, to);
    }
    return TRUE;
}

extern gboolean
matrix_disconnect(Matrix matrix,
	GIOChannel *from,
	GIOChannel *to,
	GError **error) {
    Connection *c;

    g_assert(matrix);
    g_assert(from);
    g_assert(to);
    c = g_tree_lookup(matrix->connections, from);
    if (c == NULL) {
	/* FIXME: report an error here? */
	return TRUE;
    }
    if (g_list_find(c->to, to) == NULL) {
	/* FIXME: report an error here? */
	return TRUE;
    }
    c->to = g_list_remove(c->to, to);
    /* FIXME: remove from entry from matrix if its to list is empty? */
    return TRUE;
}

static gboolean
matrix_expunge_fe(gpointer key,
	gpointer value,
	gpointer data) {
    Connection *c = value;
    GIOChannel *chan = data;

    c->to = g_list_remove(c->to, chan);
    return FALSE;
}

extern void
matrix_expunge(Matrix matrix, GIOChannel *chan) {
    g_assert(matrix);
    g_assert(chan);

    g_tree_remove(matrix->connections, chan);
    g_tree_foreach(matrix->connections, matrix_expunge_fe, chan);
}

==> conc-2.0.3.tar.gz: conc-2.0.3/common/matrix.c <==
#include "matrix.h"

#define MATRIX_BUFSIZ 2048

typedef struct _Connection {
    guint reader;
    GList *to; /* list of GIOChannel */
} Connection;

struct _Matrix {
    GTree *connections; /* tree of Connection */
};

static void
matrix_tree_free(gpointer data) {
    Connection *c = data;

    if (c->reader > 0) {
	g_source_remove(c->reader);
    }
    g_list_free(c->to);
    g_free(c);
}

static gint
matrix_compare(gconstpointer a,
	gconstpointer b,
	gpointer data) {
    return ((gint)b - (gint)a);
}

static gboolean
matrix_reader(GIOChannel *source,
	GIOCondition condition,
	gpointer data) {
    Connection *c = data;
    gchar buf[MATRIX_BUFSIZ];
    gsize bytes;
    GList *l;
    GError *e = NULL;

    switch (g_io_channel_read_chars(source, buf, MATRIX_BUFSIZ, &bytes, &e)) {
	case G_IO_STATUS_AGAIN:
	    return TRUE;
	case G_IO_STATUS_EOF:
	    return FALSE;
	case G_IO_STATUS_ERROR:
	    g_printerr("error reading for matrix (%s)\n", e->message);
	    g_clear_error(&e);
	    return FALSE;
	case G_IO_STATUS_NORMAL:
	    break;
	default:
	    return TRUE;
    }
    for (l = c->to; l; l = g_list_next(l)) {
	GIOChannel *to = l->data;
	gsize written = 0;

	switch (g_io_channel_write_chars(to, buf, bytes, &written, &e)) {
	    case G_IO_STATUS_AGAIN:
		g_printerr("warning: dropping data for matrix endpoint\n");
		break;
	    case G_IO_STATUS_EOF:
		g_printerr("warning: matrix endpoint closed\n");
		break;
	    case G_IO_STATUS_ERROR:
		g_printerr("error writing for matrix (%s)\n", e->message);
		g_clear_error(&e);
		break;
	    case G_IO_STATUS_NORMAL:
		if (written < bytes) {
		    g_printerr("warning: short write for matrix endpoint\n");
		}
		break;
	    default:
		break;
	}
	/* FIXME: disconnect broken channels? */
    }
    return TRUE;
}

extern Matrix
matrix_new(void) {
    Matrix m = g_new0(struct _Matrix, 1);

    m->connections = g_tree_new_full(matrix_compare, NULL, NULL,
	    matrix_tree_free);
    return m;
}

extern gboolean
matrix_connect(Matrix matrix,
	GIOChannel *from,
	GIOChannel *to,
	GError **error) {
    Connection *c;

    g_assert(matrix);
    g_assert(from);
    g_assert(to);
    c = g_tree_lookup(matrix->connections, from);
    if (!c) {
	c = g_new0(Connection, 1);
	c->reader = g_io_add_watch(from, G_IO_IN, matrix_reader, c);
	g_tree_insert(matrix->connections, from, c);
    }
    if (g_list_find(c->to, to) == NULL) {
	c->to = g_list_prepend(c->to, to);
    }
    g_io_channel_ref(from);
    g_io_channel_ref(to);
    return TRUE;
}

extern gboolean
matrix_disconnect(Matrix matrix,
	GIOChannel *from,
	GIOChannel *to,
	GError **error) {
    Connection *c;

    g_assert(matrix);
    g_assert(from);
    g_assert(to);
    c = g_tree_lookup(matrix->connections, from);
    if (c == NULL) {
	/* FIXME: report an error here? */
	return TRUE;
    }
    if (g_list_find(c->to, to) == NULL) {
	/* FIXME: report an error here? */
	return TRUE;
    }
    c->to = g_list_remove(c->to, to);
    g_io_channel_unref(to);
    g_io_channel_unref(from);
    if (c->to == NULL) {
	g_tree_remove(matrix->connections, from);
    }
    return TRUE;
}

struct _MatrixAndChannel {
    Matrix matrix;
    GIOChannel *chan;
    GList *matches;
};

static gboolean
matrix_expunge_fe(gpointer key,
	gpointer value,
	gpointer data) {
    GIOChannel *from = key;
    Connection *c = value;
    struct _MatrixAndChannel *m = data;

    if (g_list_find(c->to, m->chan) != NULL) {
	m->matches = g_list_prepend(m->matches, from);
    }
    return FALSE;
}

extern void
matrix_expunge(Matrix matrix, GIOChannel *chan) {
    Connection *c;
    struct _MatrixAndChannel m;

    g_assert(matrix);
    g_assert(chan);

    while ((c = g_tree_lookup(matrix->connections, chan)) != NULL &&
	    c->to != NULL) {
	matrix_disconnect(matrix, chan, (GIOChannel *)c->to->data, NULL);
    }
    m.matrix = matrix;
    m.chan = chan;
    m.matches = NULL;
    g_tree_foreach(matrix->connections, matrix_expunge_fe, &m);
    while (m.matches != NULL) {
	matrix_disconnect(matrix, m.matches->data, chan, NULL);
	m.matches = g_list_delete_link(m.matches, m.matches);
    }
}
